"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const discord_js_1 = require("discord.js");
const PropertyStore_1 = require("../../decorators/PropertyStore");
const CommandLoader_1 = require("../command/CommandLoader");
const CommandRegistry_1 = require("../command/CommandRegistry");
const EventLoader_1 = require("../event/EventLoader");
const EventRegistry_1 = require("../event/EventRegistry");
class SilentClient extends discord_js_1.Client {
    constructor(config, clientConfig) {
        super(clientConfig);
        this.mentionPrefix = false;
        this.base = true;
        this.waitForMessage = false;
        if (config.prefix)
            this.prefix = config.prefix;
        if (config.mentionPrefix !== undefined)
            this.mentionPrefix = config.mentionPrefix;
        if (!this.prefix && !this.mentionPrefix) {
            throw TypeError('You have to set a prefix.');
        }
        this.commandDir = config.commandDir;
        this.ownerID = config.owner;
        this._token = config.token;
        this.eventDir = config.eventDir;
        this.notFoundError = config.notFoundError || null;
        this.baseDir = config.baseDir || require(`${process.cwd()}/tsconfig.json`).outDir || '';
        this.base = config.base || true;
        this.waitForMessage = config.waitForMessage || false;
        this.commands = new CommandRegistry_1.CommandRegistry(this);
        this._commandLoader = new CommandLoader_1.CommandLoader(this);
        this.events = new EventRegistry_1.EventRegistry(this);
        this._eventLoader = new EventLoader_1.EventLoader(this);
        this._eventLoader.loadFromPath(this.eventDir);
        this._commandLoader.loadFromPath(this.commandDir);
        if (this.base)
            this._commandLoader.loadFromPath(null, true);
        this.checkForDuplicateCommand();
        this.on('message', this.messageHandle);
    }
    get owner() {
        return this.ownerID;
    }
    /**
     * Starts the Bot and logs into the Discord API
     */
    start() {
        // noinspection JSIgnoredPromiseFromCall
        this.login(this._token);
        return this;
    }
    getDefaultPrefix() {
        if (Array.isArray(this.prefix))
            return this.prefix[0];
        if (this.mentionPrefix)
            return this.user.toString();
        return this.prefix;
    }
    /**
     * Handle an incoming message
     * @private
     * @param msg {Message}
     */
    messageHandle(msg) {
        return __awaiter(this, void 0, void 0, function* () {
            // Run any custom message event before the commands are executed
            if (this.events.get('message')) {
                // Wait for the user defined message event and do not execute any command
                // if the user's custom event returns something falsie
                if (this.waitForMessage) {
                    let wait = this.events.get('message').run(msg);
                    if (wait.constructor.name === 'Promise')
                        wait = yield wait;
                    if (!wait)
                        return;
                }
            }
            if (msg.author.bot)
                return;
            const resolvedCommand = this.findCommand(msg);
            if (resolvedCommand === undefined)
                return;
            if (!resolvedCommand && this.notFoundError)
                return msg.reply(this.notFoundError);
            if (!resolvedCommand)
                return;
            const userRestricted = [];
            const roleRestricted = [];
            resolvedCommand.restricted.forEach((value) => {
                if (value.type === 'user')
                    userRestricted.push(value.id);
                if (value.type === 'role')
                    roleRestricted.push(value.id);
            });
            const throttle = resolvedCommand.throttle(msg.author.id);
            // Check if the user is exceeding the command rate limit
            if (throttle && throttle.uses + 1 > resolvedCommand.rateLimit.times) {
                const timeout = (throttle.start + (resolvedCommand.rateLimit.interval * 1000) - Date.now()) / 1000;
                return msg.reply(`This command is throttled for another ${timeout.toFixed(2)} seconds.`);
            }
            // Check if the user is allowed to use the command
            if (userRestricted.length && !userRestricted.includes(msg.author.id)) {
                return msg.reply('You do not have the permission to use this command');
            }
            // Check if the command is owner only and the user is the owner
            if (resolvedCommand.ownerOnly && msg.author.id !== this.owner) {
                return msg.reply('You do not have the permission to use this command');
            }
            if (msg.guild) {
                // Check if the user has the roles
                if (roleRestricted.length && msg.member.roles.filter(role => roleRestricted.includes(role.id) || roleRestricted.includes(role.name)).size === 0) {
                    return msg.reply('You do not have the permission to use this command.');
                }
                // Check for min role needed
                if (resolvedCommand.minRole) {
                    const role = msg.guild.roles.find(r => r.name.toLowerCase() === resolvedCommand.minRole.toLowerCase() ||
                        r.id === resolvedCommand.minRole);
                    if (msg.member.highestRole.comparePositionTo(role) < 0) {
                        return msg.reply('You do not have permission to use this command.');
                    }
                }
                // Check if the user has enough permissions to use the command
                if (!msg.member.permissions.has(resolvedCommand.userPermissions, false)) {
                    return msg.channel.send(`You can not use this command as you do not have all needed permissions
\`${resolvedCommand.userPermissions.join(', ')}\``);
                }
                // Check if the client has enough permissions
                if (!msg.guild.me.permissions.has(resolvedCommand.clientPermissions, false)) {
                    return msg.channel.send(`You can not use this command as the Bot needs following permissions
\`${resolvedCommand.clientPermissions.join(', ')}\``);
                }
                // Check if the command should only be used in DMs
                if (resolvedCommand.dmOnly) {
                    return msg.reply('This command can only be used in DMs.');
                }
                // Check if the command is NSFW
                if (resolvedCommand.nsfw && msg.guild && !msg.channel.nsfw) {
                    return msg.reply('This command can only be used in a NSFW marked channel.');
                }
            }
            else {
                // Check if command is set to guild only
                if (resolvedCommand.guildOnly && !msg.guild) {
                    return msg.reply('This command can only be used in a guild.');
                }
            }
            const longestPrefix = Array.isArray(this.prefix) ? this.prefix.sort((a, b) => b.length - a.length)[0] : null;
            // Parsing the argument to be an array of strings
            // Using the longest prefix for parsing
            const args = msg.content.slice(Array.isArray(this.prefix) ?
                longestPrefix.length : this.prefix.length).trim().split(/ +/g).slice(1);
            const properties = PropertyStore_1.propertyStore
                .filter(value => value.name === resolvedCommand._className.toLowerCase())
                .sort((a, b) => a.index - b.index).map(value => value.value(msg, this.getDefaultPrefix()));
            if (throttle)
                throttle.uses += 1;
            if (properties.length === 0)
                return resolvedCommand.run(msg, args);
            return resolvedCommand.run.apply(resolvedCommand, properties);
        });
    }
    findCommand(msg) {
        if (msg.isMemberMentioned(this.user) &&
            this.mentionPrefix &&
            (msg.content.startsWith(`<@${this.user.id}>`) || msg.content.startsWith(`<@!${this.user.id}>`))) {
            // tslint:disable
            const args = msg.content.split(/ +/g);
            const command = args[1];
            if (this.commands.get(command))
                return this.commands.get(command);
            if (this.commands.find(rC => rC.aliases.includes(command))) {
                return this.commands.find(rC => rC.aliases.includes(command));
            }
        }
        if (Array.isArray(this.prefix)) {
            let resolvedCommand = null;
            let loops = 0;
            for (const prefix of this.prefix) {
                if (!msg.content.startsWith(prefix))
                    continue;
                loops++;
                const args = msg.content.slice(prefix.length).trim().split(/ +/g);
                const command = args.shift().toLowerCase();
                resolvedCommand = this.commands.get(command) ||
                    this.commands.find(rC => rC.aliases.includes(command));
            }
            if (loops === 0)
                return undefined;
            return resolvedCommand;
        }
        if (!msg.content.startsWith(this.prefix))
            return undefined;
        const args = msg.content.slice(this.prefix.length).trim().split(/ +/g);
        const command = args.shift().toLowerCase();
        // tslint:enable
        const resolvedCommand = this.commands.get(command) ||
            this.commands.find(rC => rC.aliases.includes(command));
        return resolvedCommand || null;
    }
    checkForDuplicateCommand() {
        const checkedCMD = [];
        const checkedAlias = [];
        for (const { aliases, name } of this.commands.array()) {
            if (!checkedCMD.includes(name) && !checkedAlias.includes(name))
                checkedCMD.push(name);
            // tslint:disable-next-line
            else
                console.warn(`The command ${name} was initialized more than once. You probably have more than 1 command with the same name.`.toUpperCase());
            for (const alias of aliases) {
                if (!checkedAlias.includes(alias))
                    checkedAlias.push(alias);
                else
                    console.warn(`Duplicate alias declaration for command ${name}`.toUpperCase());
                if (checkedCMD.includes(alias)) {
                    console.warn(`Alias ${alias} overlaps with the ${name} command`.toUpperCase());
                }
            }
        }
    }
}
exports.SilentClient = SilentClient;
//# sourceMappingURL=SilentClient.js.map